@@ -1,6 +1,7 @@
 from dataclasses import dataclass
-from typing import List
+from typing import Any, List
 import inspect
+import itertools
 
 import requests
 
@@ -22,6 +23,8 @@ class CommandOptionType:
     USER = 6
     CHANNEL = 7
     ROLE = 8
+    MENTIONABLE = 9
+    NUMBER = 10
 
 
 class ChannelType:
@@ -35,6 +38,36 @@ class ChannelType:
     GUILD_STORE = 6
 
 
+class Permission:
+    """
+    An object representing a single permission overwrite.
+
+    ``Permission(role='1234')`` allows users with role ID 1234 to use the
+    command
+
+    ``Permission(user='5678')`` allows user ID 5678 to use the command
+
+    ``Permission(role='9012', allow=False)`` denies users with role ID 9012
+    from using the command
+    """
+
+
+    def __init__(self, role=None, user=None, allow=True):
+        if bool(role) == bool(user):
+            raise ValueError("specify only one of role or user")
+
+        self.type = 1 if role else 2
+        self.id = role or user
+        self.permission = allow
+
+    def dump(self):
+        return {
+            "type": self.type,
+            "id": self.id,
+            "permission": self.permission
+        }
+
+
 class ContextObject:
     @classmethod
     def from_dict(cls, data):
@@ -238,7 +271,11 @@ class Context(ContextObject):
     guild_id
         The unique ID (snowflake) of the guild this command was invoked in.
     options
-        An array of the options passed to the command.
+        A list of the options passed to the command.
+    values
+        A list of the values selected, if this is a Select Menu handler.
+    resolved
+        Additional data ()
     command_name
         The name of the command that was invoked.
     command_id
@@ -256,15 +293,16 @@ class Context(ContextObject):
     channel_id: str = None
     guild_id: str = None
     options: list = None
+    values: list = None
+    resolved: dict = None
     command_name: str = None
     command_id: str = None
     members: List[Member] = None
     channels: List[Channel] = None
     roles: List[Role] = None
 
-    client_id: str = ""
-    auth_headers: dict = None
-    base_url: str = ""
+    app: Any = None
+    discord: Any = None
 
     custom_id: str = None
     primary_id: str = None
@@ -275,98 +313,148 @@ class Context(ContextObject):
         if data is None:
             data = {}
 
-        custom_id = data.get("data", {}).get("custom_id")
-
         result = cls(
-            client_id = app.config["DISCORD_CLIENT_ID"] if app else "",
-            auth_headers = discord.auth_headers(app) if discord else {},
-            base_url = app.config["DISCORD_BASE_URL"] if app else "",
+            app = app,
+            discord = discord,
             author = Member.from_dict(data.get("member", {})),
             id = data.get("id"),
             token = data.get("token"),
             channel_id = data.get("channel_id"),
             guild_id = data.get("guild_id"),
             options = data.get("data", {}).get("options"),
+            values = data.get("data", {}).get("values", []),
+            resolved = data.get("resolved", {}),
             command_name = data.get("data", {}).get("name"),
             command_id = data.get("data", {}).get("id"),
-            custom_id = custom_id,
-            primary_id = custom_id.split("\n", 1)[0] if custom_id else None,
-            handler_state = custom_id.split("\n") if custom_id else None
+            custom_id = data.get("data", {}).get("custom_id") or ""
         )
 
-        result.parse_resolved(data.get("data", {}).get("resolved", {}))
+        result.parse_custom_id()
+        result.parse_resolved()
         return result
 
-    def parse_resolved(self, data):
+    @property
+    def auth_headers(self):
+        return self.discord.auth_headers(self.app)
+
+    def parse_custom_id(self):
+        """
+        Parse the custom ID of the incoming interaction data.
+
+        This includes the primary ID as well as any state stored in the
+        handler.
+        """
+
+        self.primary_id = self.custom_id.split("\n", 1)[0]
+        self.handler_state = self.custom_id.split("\n")
+
+    def parse_resolved(self):
         """
         Parse the ``"resolved"`` section of the incoming interaction data.
 
         This section includes objects representing each user, member, channel,
         and role passed as an argument to the command.
-
-        Parameters
-        ----------
-        data
-            The ``"resolved"`` section of the incoming interaction data.
         """
 
         self.members = {}
-        for id in data.get("members", {}):
-            member_info = data["members"][id]
-            member_info["user"] = data["users"][id]
+        for id in self.resolved.get("members", {}):
+            member_info = self.resolved["members"][id]
+            member_info["user"] = self.resolved["users"][id]
             self.members[id] = Member.from_dict(member_info)
 
         self.channels = {id: Channel.from_dict(data)
-                         for id, data in data.get("channels", {}).items()}
+                         for id, data
+                         in self.resolved.get("channels", {}).items()}
 
         self.roles = {id: Role.from_dict(data)
-                      for id, data in data.get("roles", {}).items()}
+                      for id, data in self.resolved.get("roles", {}).items()}
 
-    def create_args(self, data, resolved):
+    def create_args(self):
         """
         Create the arguments which will be passed to the function when the
         :class:`SlashCommand` is invoked.
+        """
 
-        This function is recursive: when a subcommand is invoked, this function
-        will call itself with the subcommand option data. This is why the
-        ``"resolved"`` data is passed as a separate argument.
+        def create_args_recursive(data, resolved):
+            if not data.get("options"):
+                return [], {}
+
+            args = []
+            kwargs = {}
+
+            for option in data["options"]:
+                if option["type"] in [
+                        CommandOptionType.SUB_COMMAND,
+                        CommandOptionType.SUB_COMMAND_GROUP]:
+
+                    args.append(option["name"])
+
+                    sub_args, sub_kwargs = create_args_recursive(
+                        option, resolved)
+
+                    args += sub_args
+                    kwargs.update(sub_kwargs)
+
+                elif option["type"] == CommandOptionType.USER:
+                    member_data = resolved["members"][option["value"]]
+                    member_data["user"] = resolved["users"][option["value"]]
+
+                    kwargs[option["name"]] = Member.from_dict(member_data)
+
+                elif option["type"] == CommandOptionType.CHANNEL:
+                    kwargs[option["name"]] = Channel.from_dict(
+                        resolved["channels"][option["value"]])
+
+                elif option["type"] == CommandOptionType.ROLE:
+                    kwargs[option["name"]] = Role.from_dict(
+                        resolved["roles"][option["value"]])
+
+                else:
+                    kwargs[option["name"]] = option["value"]
+
+            return args, kwargs
+
+        return create_args_recursive({"options": self.options}, self.resolved)
+
+    def create_handler_args(self, handler):
+        """
+        Create the arguments which will be passed to the function when a
+        custom ID handler is invoked.
 
         Parameters
         ----------
         data
             An object with the incoming data for the invocation.
-        resolved
-            The ``"resolved"`` section of the incoming interaction data.
         """
 
-        if "options" not in data:
-            return [], {}
+        args = self.handler_state[1:]
 
-        args = []
-        kwargs = {}
-        for option in data["options"]:
-            if option["type"] in [
-                    CommandOptionType.SUB_COMMAND,
-                    CommandOptionType.SUB_COMMAND_GROUP]:
-                args.append(option["name"])
-                sub_args, sub_kwargs = self.create_args(option, resolved)
-                args += sub_args
-                kwargs.update(sub_kwargs)
-            elif option["type"] == CommandOptionType.USER:
-                member_data = resolved["members"][option["value"]]
-                member_data["user"] = resolved["users"][option["value"]]
+        sig = inspect.signature(handler)
 
-                kwargs[option["name"]] = Member.from_dict(member_data)
-            elif option["type"] == CommandOptionType.CHANNEL:
-                kwargs[option["name"]] = Channel.from_dict(
-                    resolved["channels"][option["value"]])
-            elif option["type"] == CommandOptionType.ROLE:
-                kwargs[option["name"]] = Role.from_dict(
-                    resolved["roles"][option["value"]])
-            else:
-                kwargs[option["name"]] = option["value"]
+        iterator = zip(
+            itertools.count(),
+            args,
+            itertools.islice(sig.parameters.values(), 1, None)
+        )
 
-        return args, kwargs
+        for i, argument, parameter in iterator:
+            annotation = parameter.annotation
+
+            if annotation == int:
+                args[i] = int(argument)
+
+            elif annotation == bool:
+                if argument == "True":
+                    args[i] = True
+                elif argument == "False":
+                    args[i] = False
+                elif argument == "None":
+                    args[i] = None
+                else:
+                    raise ValueError(
+                        f"Invalid bool in handler state parsing: {args[i]}")
+
+        return args
 
     def followup_url(self, message=None):
         """
@@ -380,8 +468,8 @@ class Context(ContextObject):
             "@original", refers to the original message.
         """
 
-        url = (f"{self.base_url}/webhooks/"
-               f"{self.client_id}/{self.token}")
+        url = (f"{self.app.config['DISCORD_BASE_URL']}/webhooks/"
+               f"{self.app.config['DISCORD_CLIENT_ID']}/{self.token}")
         if message is not None:
             url += f"/messages/{message}"
 
@@ -444,6 +532,43 @@ class Context(ContextObject):
         response.raise_for_status()
         return response.json()["id"]
 
+    def get_command(self, command_name=None):
+        "Get the ID of a command by name."
+        if command_name is None:
+            return self.command_id
+        else:
+            try:
+                return self.app.discord_commands[command_name].id
+            except KeyError:
+                raise ValueError(f"Unknown command: {command_name}")
+
+    def overwrite_permissions(self, permissions, command=None):
+        """
+        Overwrite the permission overwrites for this command.
+
+        Parameters
+        ----------
+        permissions
+            The new list of permission overwrites.
+        command
+            The name of the command to overwrite permissions for. If omitted,
+            overwrites for the invoking command.
+        """
+
+        url = (
+            f"{self.app.config['DISCORD_BASE_URL']}/"
+            f"applications/{self.app.config['DISCORD_CLIENT_ID']}/"
+            f"guilds/{self.guild_id}/"
+            f"commands/{self.get_command(command)}/permissions"
+        )
+
+        data = [permission.dump() for permission in permissions]
+
+        response = requests.put(url, headers=self.auth_headers, json={
+            "permissions": data
+        })
+        response.raise_for_status()
+
 
 @dataclass
 class AsyncContext(Context):
@@ -455,6 +580,11 @@ class AsyncContext(Context):
     """
 
     def __post_init__(self):
+        if aiohttp == None:
+            raise ImportError(
+                "The aiohttp module is required for async usage of this "
+                "library")
+
         self.session = aiohttp.ClientSession(
             headers=self.auth_headers,
             raise_for_status=True,
@@ -512,5 +642,31 @@ class AsyncContext(Context):
         ) as response:
             return (await response.json())["id"]
 
+    async def overwrite_permissions(self, permissions, command=None):
+        """
+        Overwrite the permission overwrites for this command.
+
+        Parameters
+        ----------
+        permissions
+            The new list of permission overwrites.
+        command
+            The name of the command to overwrite permissions for. If omitted,
+            overwrites for the invoking command.
+        """
+
+        url = (
+            f"{self.app.config['DISCORD_BASE_URL']}/"
+            f"applications/{self.app.config['DISCORD_CLIENT_ID']}/"
+            f"guilds/{self.guild_id}/"
+            f"commands/{self.get_command(command)}/permissions"
+        )
+
+        data = [permission.dump() for permission in permissions]
+
+        await self.session.put(url, headers=self.auth_headers, json={
+            "permissions": data
+        })
+
     async def close(self):
-        await self.session.close()
\ No newline at end of file
+        await self.session.close()
