@@ -1,7 +1,6 @@
 import time
 import inspect
 import uuid
-import itertools
 
 import requests
 
@@ -31,8 +30,9 @@ class DiscordInteractionsBlueprint:
         self.discord_commands = {}
         self.custom_id_handlers = {}
 
-    def add_slash_command(self, command, name=None,
-                          description=None, options=None, annotations=None):
+    def add_slash_command(self, command, name=None, description=None,
+                          options=None, annotations=None,
+                          default_permission=None, permissions=None):
         """
         Create and add a new :class:`SlashCommand`.
 
@@ -50,13 +50,18 @@ class DiscordInteractionsBlueprint:
         annotations
             If ``options`` is not provided, descriptions for each of the
             options defined in the function's keyword arguments.
+        default_permission
+            Whether the command is enabled by default. Default is True.
+        permissions
+            List of permission overwrites.
         """
         slash_command = SlashCommand(
-            command, name, description, options, annotations)
+            command, name, description, options, annotations,
+            default_permission, permissions)
         self.discord_commands[slash_command.name] = slash_command
 
-    def command(self, name=None, description=None,
-                options=None, annotations=None):
+    def command(self, name=None, description=None, options=None,
+                annotations=None, default_permission=None, permissions=None):
         """
         Decorator to create a new :class:`SlashCommand`.
 
@@ -72,17 +77,22 @@ class DiscordInteractionsBlueprint:
         annotations
             If ``options`` is not provided, descriptions for each of the
             options defined in the function's keyword arguments.
+        default_permission
+            Whether the command is enabled by default. Default is True.
+        permissions
+            List of permission overwrites.
         """
 
         def decorator(func):
             nonlocal name, description, options
             self.add_slash_command(
-                func, name, description, options, annotations)
+                func, name, description, options, annotations,
+                default_permission, permissions)
             return func
 
         return decorator
 
-    def command_group(self, name, description="No description"):
+    def command_group(self, name, description="No description", is_async=False):
         """
         Create a new :class:`SlashCommandGroup`
         (which can contain multiple subcommands)
@@ -93,8 +103,12 @@ class DiscordInteractionsBlueprint:
             The name of the command group, as displayed in the Discord client.
         description
             The description of the command group.
+        is_async
+            Whether the subgroup should be considered async (if subcommands
+            get an :class:`.AsyncContext` instead of a :class:`Context`.)
         """
-        group = SlashCommandGroup(name, description)
+
+        group = SlashCommandGroup(name, description, is_async)
         self.discord_commands[name] = group
         return group
 
@@ -238,6 +252,11 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         Update the list of slash commands registered with Discord.
         This method will overwrite all existing slash commands.
 
+        Make sure you aren't calling this every time a new worker starts! You
+        will run into rate-limiting issues if multiple workers attempt to
+        register commands simultaneously. Read :ref:`workers` for more
+        info.
+
         Parameters
         ----------
         app
@@ -250,9 +269,6 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         if app is None:
             app = self.app
 
-        if app.config['DONT_REGISTER_WITH_DISCORD']:
-            return
-
         if guild_id:
             url = (f"{app.config['DISCORD_BASE_URL']}/applications/"
                    f"{app.config['DISCORD_CLIENT_ID']}/"
@@ -265,17 +281,53 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
             command.dump() for command in app.discord_commands.values()
         ]
 
-        response = requests.put(
-            url, json=overwrite_data, headers=self.auth_headers(app))
-        self.throttle(response)
+        if not app.config['DONT_REGISTER_WITH_DISCORD']:
+            response = requests.put(
+                url, json=overwrite_data, headers=self.auth_headers(app))
+            self.throttle(response)
 
-        try:
-            response.raise_for_status()
-        except requests.exceptions.HTTPError:
-            raise ValueError(
-                f"Unable to register commands:"
-                f"{response.status_code} {response.text}"
-            )
+            try:
+                response.raise_for_status()
+            except requests.exceptions.HTTPError:
+                raise ValueError(
+                    f"Unable to register commands:"
+                    f"{response.status_code} {response.text}"
+                )
+
+            for command in response.json():
+                if command["name"] in app.discord_commands:
+                    app.discord_commands[command["name"]].id = command["id"]
+        else:
+            for command in app.discord_commands.values():
+                command.id = command.name
+
+
+        url += "/permissions"
+
+        permissions_data = [
+            {
+                "id": command.id,
+                "permissions": command.dump_permissions()
+            }
+            for command in app.discord_commands.values()
+            if command.permissions
+        ]
+
+        if not permissions_data:
+            return
+
+        if not app.config['DONT_REGISTER_WITH_DISCORD']:
+            response = requests.put(
+                url, json=permissions_data, headers=self.auth_headers(app))
+            self.throttle(response)
+
+            try:
+                response.raise_for_status()
+            except requests.exceptions.HTTPError:
+                raise ValueError(
+                    f"Unable to register permissions:"
+                    f"{response.status_code} {response.text}"
+                )
 
     def throttle(self, response):
         """
@@ -354,32 +406,7 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
 
         handler = self.custom_id_handlers[context.primary_id]
 
-        args = context.handler_state[1:]
-
-        sig = inspect.signature(handler)
-
-        iterator = zip(
-            itertools.count(),
-            args,
-            itertools.islice(sig.parameters.values(), 1, None)
-        )
-
-        for i, argument, parameter in iterator:
-            annotation = parameter.annotation
-
-            if annotation == int:
-                args[i] = int(argument)
-
-            elif annotation == bool:
-                if argument == "True":
-                    args[i] = True
-                elif argument == "False":
-                    args[i] = False
-                elif argument == "None":
-                    args[i] = None
-                else:
-                    raise ValueError(
-                        f"Invalid bool in handler state parsing: {args[i]}")
+        args = context.create_handler_args(handler)
 
         result = handler(context, *args)
 
